import logging
from typing import Any, Callable

from ..flows.base import Flow
from ..samples import Samples
from ..transforms import IdentityTransform
from ..utils import track_calls

logger = logging.getLogger(__name__)


class Sampler:
    """Base class for all samplers.

    Parameters
    ----------
    log_likelihood : Callable
        The log likelihood function.
    log_prior : Callable
        The log prior function.
    dims : int
        The number of dimensions.
    flow : Flow
        The flow object.
    xp : Callable
        The array backend to use.
    parameters : list[str] | None
        The list of parameter names. If None, any samples objects will not
        have the parameters names specified.
    """

    def __init__(
        self,
        log_likelihood: Callable,
        log_prior: Callable,
        dims: int,
        prior_flow: Flow,
        xp: Callable,
        dtype: Any | str | None = None,
        parameters: list[str] | None = None,
        preconditioning_transform: Callable | None = None,
    ):
        self.prior_flow = prior_flow
        self._log_likelihood = log_likelihood
        self.log_prior = log_prior
        self.dims = dims
        self.xp = xp
        self.dtype = dtype
        self.parameters = parameters
        self.history = None
        self.n_likelihood_evaluations = 0
        if preconditioning_transform is None:
            self.preconditioning_transform = IdentityTransform(xp=self.xp)
        else:
            self.preconditioning_transform = preconditioning_transform

    def fit_preconditioning_transform(self, x):
        """Fit the data transform to the data."""
        x = self.preconditioning_transform.xp.asarray(x)
        return self.preconditioning_transform.fit(x)

    @track_calls
    def sample(self, n_samples: int) -> Samples:
        raise NotImplementedError

    def log_likelihood(self, samples: Samples) -> Samples:
        """Computes the log likelihood of the samples.

        Also tracks the number of likelihood evaluations.
        """
        self.n_likelihood_evaluations += len(samples)
        return self._log_likelihood(samples)

    def config_dict(self, include_sample_calls: bool = True) -> dict:
        """
        Returns a dictionary with the configuration of the sampler.

        Parameters
        ----------
        include_sample_calls : bool
            Whether to include the sample calls in the configuration.
            Default is True.
        """
        config = {}
        if include_sample_calls:
            if hasattr(self, "sample") and hasattr(self.sample, "calls"):
                config["sample_calls"] = self.sample.calls.to_dict(
                    list_to_dict=True
                )
            else:
                logger.warning(
                    "Sampler does not have a sample method with calls attribute."
                )
        return config



import logging
from typing import Callable

import jax
import jax.numpy as jnp
import jax.random as jrandom
from flowjax.train import fit_to_data

from ...transforms import IdentityTransform
from ...utils import decode_dtype, encode_dtype, resolve_dtype
from ..base import Flow
from .utils import get_flow

logger = logging.getLogger(__name__)


class FlowJax(Flow):
    xp = jnp

    def __init__(
        self,
        dims: int,
        key=None,
        data_transform=None,
        dtype=None,
        **kwargs,
    ):
        device = kwargs.pop("device", None)
        if device is not None:
            logger.warning("The device argument is not used in FlowJax. ")
        resolved_dtype = (
            resolve_dtype(dtype, jnp)
            if dtype is not None
            else jnp.dtype(jnp.float32)
        )
        if data_transform is None:
            data_transform = IdentityTransform(self.xp, dtype=resolved_dtype)
        elif getattr(data_transform, "dtype", None) is None:
            data_transform.dtype = resolved_dtype
        super().__init__(dims, device=device, data_transform=data_transform)
        self.dtype = resolved_dtype
        if key is None:
            key = jrandom.key(0)
            logger.warning(
                "The key argument is None. "
                "A random key will be used for the flow. "
                "Results may not be reproducible."
            )
        self.key = key
        self.loc = None
        self.scale = None
        self.key, subkey = jrandom.split(self.key)
        self._flow = get_flow(
            key=subkey,
            dims=self.dims,
            dtype=self.dtype,
            **kwargs,
        )

    def fit(self, x, **kwargs):
        from ...history import FlowHistory

        x = jnp.asarray(x, dtype=self.dtype)
        x_prime = jnp.asarray(self.fit_data_transform(x), dtype=self.dtype)
        self.key, subkey = jrandom.split(self.key)
        self._flow, losses = fit_to_data(subkey, self._flow, x_prime, **kwargs)
        return FlowHistory(
            training_loss=list(losses["train"]),
            validation_loss=list(losses["val"]),
        )

    def forward(self, x, xp: Callable = jnp):
        x = jnp.asarray(x, dtype=self.dtype)
        x_prime, log_abs_det_jacobian = self.rescale(x)
        x_prime = jnp.asarray(x_prime, dtype=self.dtype)
        z, log_abs_det_jacobian_flow = self._flow.forward(x_prime)
        return xp.asarray(z), xp.asarray(
            log_abs_det_jacobian + log_abs_det_jacobian_flow
        )

    def inverse(self, z, xp: Callable = jnp):
        z = jnp.asarray(z, dtype=self.dtype)
        x_prime, log_abs_det_jacobian_flow = self._flow.inverse(z)
        x_prime = jnp.asarray(x_prime, dtype=self.dtype)
        x, log_abs_det_jacobian = self.inverse_rescale(x_prime)
        return xp.asarray(x), xp.asarray(
            log_abs_det_jacobian + log_abs_det_jacobian_flow
        )

    def log_prob(self, x, xp: Callable = jnp):
        x = jnp.asarray(x, dtype=self.dtype)
        x_prime, log_abs_det_jacobian = self.rescale(x)
        x_prime = jnp.asarray(x_prime, dtype=self.dtype)
        log_prob = self._flow.log_prob(x_prime)
        return xp.asarray(log_prob + log_abs_det_jacobian)

    def sample(self, n_samples: int, xp: Callable = jnp):
        self.key, subkey = jrandom.split(self.key)
        x_prime = self._flow.sample(subkey, (n_samples,))
        x = self.inverse_rescale(x_prime)[0]
        return xp.asarray(x)

    def sample_and_log_prob(self, n_samples: int, xp: Callable = jnp):
        self.key, subkey = jrandom.split(self.key)
        x_prime = self._flow.sample(subkey, (n_samples,))
        log_prob = self._flow.log_prob(x_prime)
        x, log_abs_det_jacobian = self.inverse_rescale(x_prime)
        return xp.asarray(x), xp.asarray(log_prob - log_abs_det_jacobian)

    def save(self, h5_file, path="flow"):
        import equinox as eqx
        from array_api_compat import numpy as np

        from ...utils import recursively_save_to_h5_file

        grp = h5_file.require_group(path)

        # ---- config ----
        config = self.config_dict().copy()
        config.pop("key", None)
        config["key_data"] = jax.random.key_data(self.key)
        dtype_value = config.get("dtype")
        if dtype_value is None:
            dtype_value = self.dtype
        else:
            dtype_value = jnp.dtype(dtype_value)
        print(dtype_value)
        config["dtype"] = encode_dtype(jnp, dtype_value)

        data_transform = config.pop("data_transform", None)
        if data_transform is not None:
            data_transform.save(grp, "data_transform")

        recursively_save_to_h5_file(grp, "config", config)

        # ---- save arrays ----
        arrays, _ = eqx.partition(self._flow, eqx.is_array)
        leaves, _ = jax.tree_util.tree_flatten(arrays)

        params_grp = grp.require_group("params")
        # clear old datasets
        for name in list(params_grp.keys()):
            del params_grp[name]

        for i, p in enumerate(leaves):
            params_grp.create_dataset(str(i), data=np.asarray(p))

    @classmethod
    def load(cls, h5_file, path="flow"):
        import equinox as eqx

        from ...utils import load_from_h5_file

        grp = h5_file[path]

        # ---- config ----
        config = load_from_h5_file(grp, "config")
        config["dtype"] = decode_dtype(jnp, config.get("dtype"))
        if "data_transform" in grp:
            from ...transforms import BaseTransform

            config["data_transform"] = BaseTransform.load(
                grp,
                "data_transform",
                strict=False,
            )

        key_data = config.pop("key_data", None)
        if key_data is not None:
            config["key"] = jax.random.wrap_key_data(key_data)

        kwargs = config.pop("kwargs", {})
        config.update(kwargs)

        # build object (will replace its _flow)
        obj = cls(**config)

        # ---- load arrays ----
        params_grp = grp["params"]
        loaded_params = [
            jnp.array(params_grp[str(i)][:]) for i in range(len(params_grp))
        ]

        # rebuild template flow
        kwargs.pop("device")
        flow_template = get_flow(key=jrandom.key(0), dims=obj.dims, **kwargs)
        arrays_template, static = eqx.partition(flow_template, eqx.is_array)

        # use treedef from template
        treedef = jax.tree_util.tree_structure(arrays_template)
        arrays = jax.tree_util.tree_unflatten(treedef, loaded_params)

        # recombine
        obj._flow = eqx.combine(static, arrays)

        return obj



import copy
import logging
from typing import Callable

import array_api_compat.torch as torch_api
import torch
import tqdm
import zuko
from array_api_compat import is_numpy_namespace, is_torch_array

from ...history import FlowHistory
from ...transforms import IdentityTransform
from ...utils import decode_dtype, encode_dtype, resolve_dtype
from ..base import Flow

logger = logging.getLogger(__name__)


class BaseTorchFlow(Flow):
    _flow = None
    xp = torch_api

    def __init__(
        self,
        dims: int,
        seed: int = 1234,
        device: str = "cpu",
        data_transform=None,
        dtype=None,
    ):
        resolved_dtype = (
            resolve_dtype(dtype, torch)
            if dtype is not None
            else torch.get_default_dtype()
        )
        if data_transform is None:
            data_transform = IdentityTransform(self.xp, dtype=resolved_dtype)
        elif getattr(data_transform, "dtype", None) is None:
            data_transform.dtype = resolved_dtype
        super().__init__(
            dims,
            device=torch.device(device or "cpu"),
            data_transform=data_transform,
        )
        self.dtype = resolved_dtype
        torch.manual_seed(seed)
        self.loc = None
        self.scale = None

    @property
    def flow(self):
        return self._flow

    @flow.setter
    def flow(self, flow):
        self._flow = flow
        self._flow.to(device=self.device, dtype=self.dtype)
        self._flow.compile()

    def fit(self, x) -> FlowHistory:
        raise NotImplementedError()

    def save(self, h5_file, path="flow"):
        """Save the weights of the flow to an HDF5 file."""
        from ...utils import recursively_save_to_h5_file

        flow_grp = h5_file.create_group(path)
        # Save config
        config = self.config_dict().copy()
        data_transform = config.pop("data_transform", None)
        dtype_value = config.get("dtype")
        if dtype_value is None:
            dtype_value = self.dtype
        else:
            dtype_value = resolve_dtype(dtype_value, torch)
        config["dtype"] = encode_dtype(torch, dtype_value)
        if data_transform is not None:
            data_transform.save(flow_grp, "data_transform")
        recursively_save_to_h5_file(flow_grp, "config", config)
        # Save weights
        weights_grp = flow_grp.create_group("weights")
        for name, tensor in self._flow.state_dict().items():
            weights_grp.create_dataset(name, data=tensor.cpu().numpy())

    @classmethod
    def load(self, h5_file, path="flow"):
        """Load the weights of the flow from an HDF5 file."""
        from ...utils import load_from_h5_file

        flow_grp = h5_file[path]
        # Load config
        config = load_from_h5_file(flow_grp, "config")
        config["dtype"] = decode_dtype(torch, config.get("dtype"))
        if "data_transform" in flow_grp:
            from ..transforms import BaseTransform

            data_transform = BaseTransform.load(
                flow_grp,
                "data_transform",
                strict=False,
            )
            config["data_transform"] = data_transform
        obj = self(**config)
        # Load weights
        weights = {
            name: torch.tensor(data[()])
            for name, data in flow_grp["weights"].items()
        }
        obj._flow.load_state_dict(weights)
        return obj


class ZukoFlow(BaseTorchFlow):
    def __init__(
        self,
        dims,
        flow_class: str | Callable = "MAF",
        data_transform=None,
        seed=1234,
        device: str = "cpu",
        dtype=None,
        **kwargs,
    ):
        super().__init__(
            dims,
            device=device,
            data_transform=data_transform,
            seed=seed,
            dtype=dtype,
        )

        if isinstance(flow_class, str):
            FlowClass = getattr(zuko.flows, flow_class)
        else:
            FlowClass = flow_class

        # Ints are some times passed as strings, so we convert them
        if hidden_features := kwargs.pop("hidden_features", None):
            kwargs["hidden_features"] = list(map(int, hidden_features))

        self.flow = FlowClass(self.dims, 0, **kwargs)
        logger.info(f"Initialized normalizing flow: \n {self.flow}\n")

    def loss_fn(self, x):
        return -self.flow().log_prob(x).mean()

    def fit(
        self,
        x,
        n_epochs: int = 100,
        lr: float = 1e-3,
        batch_size: int = 500,
        validation_fraction: float = 0.2,
        clip_grad: float | None = None,
        lr_annealing: bool = False,
    ):
        from ...history import FlowHistory

        if not is_torch_array(x):
            x = torch.tensor(x, dtype=self.dtype, device=self.device)
        else:
            x = torch.clone(x)
            x = x.type(self.dtype)
            x = x.to(self.device)
        x_prime = self.fit_data_transform(x)
        indices = torch.randperm(x_prime.shape[0])
        x_prime = x_prime[indices, ...]

        n = x_prime.shape[0]
        x_train = torch.as_tensor(
            x_prime[: -int(validation_fraction * n)],
            dtype=self.dtype,
            device=self.device,
        )

        logger.info(
            f"Training on {x_train.shape[0]} samples, "
            f"validating on {x_prime.shape[0] - x_train.shape[0]} samples."
        )

        if torch.isnan(x_train).any():
            dims_with_nan = (
                torch.isnan(x_train).any(dim=0).nonzero(as_tuple=True)[0]
            )
            raise ValueError(
                f"Training data contains NaN values in dimensions: {dims_with_nan.tolist()}"
            )
        if not torch.isfinite(x_train).all():
            dims_with_inf = (
                (~torch.isfinite(x_train)).any(dim=0).nonzero(as_tuple=True)[0]
            )
            raise ValueError(
                f"Training data contains infinite values in dimensions: {dims_with_inf.tolist()}"
            )

        x_val = torch.as_tensor(
            x_prime[-int(validation_fraction * n) :],
            dtype=self.dtype,
            device=self.device,
        )
        if torch.isnan(x_val).any():
            raise ValueError("Validation data contains infinite values.")

        if not torch.isfinite(x_val).all():
            raise ValueError("Validation data contains infinite values.")

        dataset = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(x_train),
            shuffle=True,
            batch_size=batch_size,
        )
        val_dataset = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(x_val),
            shuffle=False,
            batch_size=batch_size,
        )

        # Train to maximize the log-likelihood
        optimizer = torch.optim.Adam(self._flow.parameters(), lr=lr)
        if lr_annealing:
            scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                optimizer, n_epochs
            )
        history = FlowHistory()

        best_val_loss = float("inf")
        best_flow_state = None

        with tqdm.tqdm(range(n_epochs), desc="Epochs") as pbar:
            for _ in pbar:
                self.flow.train()
                loss_epoch = 0.0
                for (x_batch,) in dataset:
                    loss = self.loss_fn(x_batch)
                    optimizer.zero_grad()
                    loss.backward()
                    if clip_grad is not None:
                        torch.nn.utils.clip_grad_norm_(
                            self.flow.parameters(), clip_grad
                        )
                    optimizer.step()
                    loss_epoch += loss.item()
                if lr_annealing:
                    scheduler.step()
                avg_train_loss = loss_epoch / len(dataset)
                history.training_loss.append(avg_train_loss)
                self.flow.eval()
                val_loss = 0.0
                for (x_batch,) in val_dataset:
                    with torch.no_grad():
                        val_loss += self.loss_fn(x_batch).item()
                avg_val_loss = val_loss / len(val_dataset)
                if avg_val_loss < best_val_loss:
                    best_val_loss = avg_val_loss
                    best_flow_state = copy.deepcopy(self.flow.state_dict())

                history.validation_loss.append(avg_val_loss)
                pbar.set_postfix(
                    train_loss=f"{avg_train_loss:.4f}",
                    val_loss=f"{avg_val_loss:.4f}",
                )
        if best_flow_state is not None:
            self.flow.load_state_dict(best_flow_state)
            logger.info(f"Loaded best model with val loss {best_val_loss:.4f}")

        self.flow.eval()
        return history

    def sample_and_log_prob(self, n_samples: int, xp=torch_api):
        with torch.no_grad():
            x_prime, log_prob = self.flow().rsample_and_log_prob((n_samples,))
        x, log_abs_det_jacobian = self.inverse_rescale(x_prime)
        return xp.asarray(x), xp.asarray(log_prob - log_abs_det_jacobian)

    def sample(self, n_samples: int, xp=torch_api):
        with torch.no_grad():
            x_prime = self.flow().rsample((n_samples,))
        x = self.inverse_rescale(x_prime)[0]
        return xp.asarray(x)

    def log_prob(self, x, xp=torch_api):
        x = torch.as_tensor(x, dtype=self.dtype, device=self.device)
        x_prime, log_abs_det_jacobian = self.rescale(x)
        return xp.asarray(
            self._flow().log_prob(x_prime) + log_abs_det_jacobian
        )

    def forward(self, x, xp=torch_api):
        x = torch.as_tensor(x, dtype=self.dtype, device=self.device)
        x_prime, log_j_rescale = self.rescale(x)
        z, log_abs_det_jacobian = self._flow().transform.call_and_ladj(x_prime)
        if is_numpy_namespace(xp):
            # Convert to numpy namespace if needed
            z = z.detach().numpy()
            log_abs_det_jacobian = log_abs_det_jacobian.detach().numpy()
            log_j_rescale = log_j_rescale.detach().numpy()
        return xp.asarray(z), xp.asarray(log_abs_det_jacobian + log_j_rescale)

    def inverse(self, z, xp=torch_api):
        z = torch.as_tensor(z, dtype=self.dtype, device=self.device)
        with torch.no_grad():
            x_prime, log_abs_det_jacobian = (
                self._flow().transform.inv.call_and_ladj(z)
            )
        x, log_j_rescale = self.inverse_rescale(x_prime)
        if is_numpy_namespace(xp):
            # Convert to numpy namespace if needed
            x = x.detach().numpy()
            log_abs_det_jacobian = log_abs_det_jacobian.detach().numpy()
            log_j_rescale = log_j_rescale.detach().numpy()
        return xp.asarray(x), xp.asarray(log_j_rescale + log_abs_det_jacobian)


class ZukoFlowMatching(ZukoFlow):
    def __init__(
        self,
        dims,
        data_transform=None,
        seed=1234,
        device="cpu",
        eta: float = 1e-3,
        dtype=None,
        **kwargs,
    ):
        kwargs.setdefault("hidden_features", 4 * [100])
        super().__init__(
            dims,
            seed=seed,
            device=device,
            data_transform=data_transform,
            flow_class="CNF",
            dtype=dtype,
        )
        self.eta = eta

    def loss_fn(self, theta: torch.Tensor):
        t = torch.rand(
            theta.shape[:-1], dtype=theta.dtype, device=theta.device
        )
        t_ = t[..., None]
        eps = torch.randn_like(theta)
        theta_prime = (1 - t_) * theta + (t_ + self.eta) * eps
        v = eps - theta
        return (self._flow.transform.f(t, theta_prime) - v).square().mean()



import inspect
import logging
from typing import Any

from ..history import FlowHistory
from ..transforms import BaseTransform, IdentityTransform

logger = logging.getLogger(__name__)


class Flow:
    xp = None  # type: Any

    def __init__(
        self,
        dims: int,
        device: Any,
        data_transform: BaseTransform | None = None,
    ):
        self.dims = dims
        self.device = device

        if data_transform is None:
            data_transform = IdentityTransform(self.xp)
            logger.info("No data_transform provided, using IdentityTransform.")

        self.data_transform = data_transform

    def log_prob(self, x):
        raise NotImplementedError

    def sample(self, x):
        raise NotImplementedError

    def sample_and_log_prob(self, n_samples):
        raise NotImplementedError

    def fit(self, samples, **kwargs) -> FlowHistory:
        raise NotImplementedError

    def fit_data_transform(self, x):
        return self.data_transform.fit(x)

    def rescale(self, x):
        return self.data_transform.forward(x)

    def inverse_rescale(self, x):
        return self.data_transform.inverse(x)

    def config_dict(self):
        """Return a dictionary of the configuration of the flow.

        This can be used to recreate the flow by passing the dictionary
        as keyword arguments to the constructor.

        This is automatically populated with the arguments passed to the
        constructor.

        Returns
        -------
        config : dict
            The configuration dictionary.
        """
        return getattr(self, "_init_args", {})

    def save(self, h5_file, path="flow"):
        raise NotImplementedError

    @classmethod
    def load(cls, h5_file, path="flow"):
        raise NotImplementedError

    def __new__(cls, *args, **kwargs):
        # Create instance
        obj = super().__new__(cls)
        # Inspect the subclass's __init__ signature
        sig = inspect.signature(cls.__init__)
        bound = sig.bind_partial(obj, *args, **kwargs)
        bound.apply_defaults()
        # Save args (excluding self)
        obj._init_args = {
            k: v for k, v in bound.arguments.items() if k != "self"
        }
        return obj



from functools import partial

import numpy as np

from ...samples import SMCSamples
from ...utils import to_numpy, track_calls
from .base import NumpySMCSampler


class MiniPCNSMC(NumpySMCSampler):
    """MiniPCN SMC sampler."""

    rng = None

    def log_prob(self, x, beta=None):
        return to_numpy(super().log_prob(x, beta))

    @track_calls
    def sample(
        self,
        n_samples: int,
        n_steps: int = None,
        min_step: float | None = None,
        max_n_steps: int | None = None,
        adaptive: bool = True,
        target_efficiency: float = 0.5,
        target_efficiency_rate: float = 1.0,
        n_final_samples: int | None = None,
        sampler_kwargs: dict | None = None,
        rng: np.random.Generator | None = None,
    ):
        self.sampler_kwargs = sampler_kwargs or {}
        self.sampler_kwargs.setdefault("n_steps", 5 * self.dims)
        self.sampler_kwargs.setdefault("target_acceptance_rate", 0.234)
        self.sampler_kwargs.setdefault("step_fn", "tpcn")
        self.rng = rng or np.random.default_rng()
        return super().sample(
            n_samples,
            n_steps=n_steps,
            adaptive=adaptive,
            target_efficiency=target_efficiency,
            target_efficiency_rate=target_efficiency_rate,
            n_final_samples=n_final_samples,
            min_step=min_step,
            max_n_steps=max_n_steps,
        )

    def mutate(self, particles, beta, n_steps=None):
        from minipcn import Sampler

        log_prob_fn = partial(self.log_prob, beta=beta)

        sampler = Sampler(
            log_prob_fn=log_prob_fn,
            step_fn=self.sampler_kwargs["step_fn"],
            rng=self.rng,
            dims=self.dims,
            target_acceptance_rate=self.sampler_kwargs[
                "target_acceptance_rate"
            ],
        )
        # Map to transformed dimension for sampling
        z = to_numpy(self.fit_preconditioning_transform(particles.x))
        chain, history = sampler.sample(
            z,
            n_steps=n_steps or self.sampler_kwargs["n_steps"],
        )
        x = self.preconditioning_transform.inverse(chain[-1])[0]

        self.history.mcmc_acceptance.append(np.mean(history.acceptance_rate))

        samples = SMCSamples(x, xp=self.xp, beta=beta, dtype=self.dtype)
        samples.log_q = samples.array_to_namespace(
            self.prior_flow.log_prob(samples.x)
        )
        samples.log_prior = samples.array_to_namespace(self.log_prior(samples))
        samples.log_likelihood = samples.array_to_namespace(
            self.log_likelihood(samples)
        )
        if samples.xp.isnan(samples.log_q).any():
            raise ValueError("Log proposal contains NaN values")
        return samples



import pytest

from aspire import Aspire
from aspire.utils import AspireFile


@pytest.fixture(params=[None, "float32", "float64"])
def dtype(request):
    return request.param


def test_integration_zuko(
    log_likelihood,
    log_prior,
    dims,
    samples,
    parameters,
    prior_bounds,
    bounded_to_unbounded,
    sampler_config,
    dtype,
    tmp_path,
):
    if "blackjax_smc" in sampler_config.sampler:
        pytest.xfail(reason="BlackJAX requires JAX arrays.")

    aspire = Aspire(
        log_likelihood=log_likelihood,
        log_prior=log_prior,
        dims=dims,
        parameters=parameters,
        prior_bounds=prior_bounds,
        flow_matching=False,
        bounded_to_unbounded=bounded_to_unbounded,
        flow_backend="zuko",
        dtype=dtype,
    )
    aspire.fit(samples, n_epochs=5)
    aspire.sample_posterior(
        n_samples=100,
        sampler=sampler_config.sampler,
        **sampler_config.sampler_kwargs,
    )
    with AspireFile(tmp_path / "test_integration_zuko.h5", "w") as h5_file:
        aspire.save_config(h5_file)
        aspire.save_flow(h5_file)
        samples.save(h5_file, path="posterior_samples")


@pytest.mark.requires("flowjax")
def test_integration_flowjax(
    log_likelihood,
    log_prior,
    dims,
    samples,
    parameters,
    prior_bounds,
    bounded_to_unbounded,
    samples_backend,
    sampler_config,
    dtype,
    tmp_path,
):
    import jax

    if "blackjax_smc" in sampler_config.sampler:
        if samples_backend != "jax":
            pytest.xfail(reason="BlackJAX requires JAX arrays.")
        if dtype == "float32":
            pytest.xfail(
                reason="BlackJAX tests with float32 fail when running with jax defaults set to float64 dtypes."
            )

    aspire = Aspire(
        log_likelihood=log_likelihood,
        log_prior=log_prior,
        dims=dims,
        parameters=parameters,
        prior_bounds=prior_bounds,
        flow_matching=False,
        bounded_to_unbounded=bounded_to_unbounded,
        flow_backend="flowjax",
        key=jax.random.key(0),
        dtype=dtype,
    )
    aspire.fit(samples, max_epochs=5)
    posterior_samples = aspire.sample_posterior(
        n_samples=100,
        sampler=sampler_config.sampler,
        **sampler_config.sampler_kwargs,
    )

    with AspireFile(tmp_path / "test_integration_flowjax.h5", "w") as h5_file:
        aspire.save_config(h5_file)
        aspire.save_flow(h5_file)
        posterior_samples.save(h5_file, path="posterior_samples")


def test_init_existing_flow(
    log_likelihood,
    log_prior,
    dims,
    parameters,
    prior_bounds,
    bounded_to_unbounded,
):
    aspire_kwargs = {
        "log_likelihood": log_likelihood,
        "log_prior": log_prior,
        "dims": dims,
        "parameters": parameters,
        "prior_bounds": prior_bounds,
        "flow_matching": False,
        "bounded_to_unbounded": bounded_to_unbounded,
        "flow_backend": "zuko",
    }

    aspire = Aspire(**aspire_kwargs)
    aspire.init_flow()

    saved_flow = aspire.flow
    new_aspire_obj = Aspire(**aspire_kwargs | {"flow": saved_flow})

    assert new_aspire_obj.flow is aspire.flow

    new_aspire_obj.flow = saved_flow

    assert new_aspire_obj.flow is saved_flow



from __future__ import annotations

import inspect
import logging
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Any

import array_api_compat.numpy as np
import h5py
import wrapt
from array_api_compat import (
    array_namespace,
    is_jax_array,
    is_torch_array,
    is_torch_namespace,
    to_device,
)

if TYPE_CHECKING:
    from multiprocessing import Pool

    from array_api_compat.common._typing import Array

    from .aspire import Aspire

logger = logging.getLogger(__name__)


def configure_logger(
    log_level: str | int = "INFO",
    additional_loggers: list[str] = None,
    include_aspire_loggers: bool = True,
) -> logging.Logger:
    """Configure the logger.

    Adds a stream handler to the logger.

    Parameters
    ----------
    log_level : str or int, optional
        The log level to use. Defaults to "INFO".
    additional_loggers : list of str, optional
        Additional loggers to configure. Defaults to None.
    include_aspire_loggers : bool, optional
        Whether to include all loggers that start with "aspire_" or "aspire-".
        Defaults to True.

    Returns
    -------
    logging.Logger
        The configured logger.
    """
    logger = logging.getLogger("aspire")
    logger.setLevel(log_level)
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    formatter = logging.Formatter(
        "%(asctime)s - aspire - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    additional_loggers = additional_loggers or []
    for name in logger.manager.loggerDict:
        if include_aspire_loggers and (
            name.startswith("aspire_") or name.startswith("aspire-")
        ):
            additional_loggers.append(name)

    for name in additional_loggers:
        dep_logger = logging.getLogger(name)
        dep_logger.setLevel(log_level)
        dep_logger.handlers.clear()
        for handler in logger.handlers:
            dep_logger.addHandler(handler)
        dep_logger.propagate = False

    return logger


class PoolHandler:
    """Context manager to temporarily replace the log_likelihood method of a
    aspire instance with a version that uses a multiprocessing pool to
    parallelize computation.

    Parameters
    ----------
    aspire_instance : aspire
        The aspire instance to modify. The log_likelihood method of this
        instance must accept a :code:`map_fn` keyword argument.
    pool : multiprocessing.Pool
        The pool to use for parallel computation.
    close_pool : bool, optional
        Whether to close the pool when exiting the context manager.
        Defaults to True.
    parallelize_prior : bool, optional
        Whether to parallelize the log_prior method as well. Defaults to False.
        If True, the log_prior method of the aspire instance must also
        accept a :code:`map_fn` keyword argument.
    """

    def __init__(
        self,
        aspire_instance: Aspire,
        pool: Pool,
        close_pool: bool = True,
        parallelize_prior: bool = False,
    ):
        self.parallelize_prior = parallelize_prior
        self.aspire_instance = aspire_instance
        self.pool = pool
        self.close_pool = close_pool

    @property
    def aspire_instance(self):
        return self._aspire_instance

    @aspire_instance.setter
    def aspire_instance(self, value: Aspire):
        signature = inspect.signature(value.log_likelihood)
        if "map_fn" not in signature.parameters:
            raise ValueError(
                "The log_likelihood method of the Aspire instance must accept a"
                " 'map_fn' keyword argument."
            )
        signature = inspect.signature(value.log_prior)
        if "map_fn" not in signature.parameters and self.parallelize_prior:
            raise ValueError(
                "The log_prior method of the Aspire instance must accept a"
                " 'map_fn' keyword argument if parallelize_prior is True."
            )
        self._aspire_instance = value

    def __enter__(self):
        self.original_log_likelihood = self.aspire_instance.log_likelihood
        self.original_log_prior = self.aspire_instance.log_prior
        if self.pool is not None:
            logger.debug("Updating map function in log-likelihood method")
            self.aspire_instance.log_likelihood = partial(
                self.original_log_likelihood, map_fn=self.pool.map
            )
            if self.parallelize_prior:
                logger.debug("Updating map function in log-prior method")
                self.aspire_instance.log_prior = partial(
                    self.original_log_prior, map_fn=self.pool.map
                )
        return self.pool

    def __exit__(self, exc_type, exc_value, traceback):
        self.aspire_instance.log_likelihood = self.original_log_likelihood
        self.aspire_instance.log_prior = self.original_log_prior
        if self.close_pool:
            logger.debug("Closing pool")
            self.pool.close()
            self.pool.join()
        else:
            logger.debug("Not closing pool")


def logit(x: Array, eps: float | None = None) -> tuple[Array, Array]:
    """Logit function that also returns log Jacobian determinant.

    Parameters
    ----------
    x : float or ndarray
        Array of values
    eps : float, optional
        Epsilon value used to clamp inputs to [eps, 1 - eps]. If None, then
        inputs are not clamped.

    Returns
    -------
    float or ndarray
        Rescaled values.
    float or ndarray
        Log Jacobian determinant.
    """
    xp = array_namespace(x)
    if eps:
        x = xp.clip(x, eps, 1 - eps)
    y = xp.log(x) - xp.log1p(-x)
    log_j = (-xp.log(x) - xp.log1p(-x)).sum(-1)
    return y, log_j


def sigmoid(x: Array) -> tuple[Array, Array]:
    """Sigmoid function that also returns log Jacobian determinant.

    Parameters
    ----------
    x : float or ndarray
        Array of values

    Returns
    -------
    float or ndarray
        Rescaled values.
    float or ndarray
        Log Jacobian determinant.
    """
    xp = array_namespace(x)
    x = xp.divide(1, 1 + xp.exp(-x))
    log_j = (xp.log(x) + xp.log1p(-x)).sum(-1)
    return x, log_j


def logsumexp(x: Array, axis: int | None = None) -> Array:
    """Implementation of logsumexp that works with array api.

    This will be removed once the implementation in scipy is compatible.
    """
    xp = array_namespace(x)
    c = x.max()
    return c + xp.log(xp.sum(xp.exp(x - c), axis=axis))


def to_numpy(x: Array, **kwargs) -> np.ndarray:
    """Convert an array to a numpy array.

    This automatically moves the array to the CPU.

    Parameters
    ----------
    x : Array
        The array to convert.
    kwargs : dict
        Additional keyword arguments to pass to numpy.asarray.
    """
    try:
        return np.asarray(to_device(x, "cpu"), **kwargs)
    except (ValueError, NotImplementedError):
        return np.asarray(x, **kwargs)


def asarray(x, xp: Any = None, **kwargs) -> Array:
    """Convert an array to the specified array API.

    Parameters
    ----------
    x : Array
        The array to convert.
    xp : Any
        The array API to use for the conversion. If None, the array API
        is inferred from the input array.
    kwargs : dict
        Additional keyword arguments to pass to xp.asarray.
    """
    if is_jax_array(x) and is_torch_namespace(xp):
        return xp.utils.dlpack.from_dlpack(x)
    else:
        return xp.asarray(x, **kwargs)


def resolve_dtype(dtype: Any | str | None, xp: Any) -> Any | None:
    """Resolve a dtype specification into an XP-specific dtype.

    Parameters
    ----------
    dtype : Any | str | None
        The dtype specification. Can be None, a string, or a dtype-like object.
    xp : module
        The array API module that should interpret the dtype.

    Returns
    -------
    Any | None
        The resolved dtype object compatible with ``xp`` (or None if unspecified).
    """
    if dtype is None or xp is None:
        return dtype

    if isinstance(dtype, str):
        dtype_name = _dtype_to_name(dtype)
        if is_torch_namespace(xp):
            resolved = getattr(xp, dtype_name, None)
            if resolved is None:
                raise ValueError(
                    f"Unknown dtype '{dtype}' for namespace {xp.__name__}"
                )
            return resolved
        try:
            return xp.dtype(dtype_name)
        except (AttributeError, TypeError, ValueError):
            resolved = getattr(xp, dtype_name, None)
            if resolved is not None:
                return resolved
            raise ValueError(
                f"Unknown dtype '{dtype}' for namespace {getattr(xp, '__name__', xp)}"
            )

    if is_torch_namespace(xp):
        return dtype

    try:
        return xp.dtype(dtype)
    except (AttributeError, TypeError, ValueError):
        return dtype


def _dtype_to_name(dtype: Any | str | None) -> str | None:
    """Extract a canonical (lowercase) name for a dtype-like object."""
    if dtype is None:
        return None
    if isinstance(dtype, str):
        name = dtype
    elif hasattr(dtype, "name") and getattr(dtype, "name"):
        name = dtype.name
    elif hasattr(dtype, "__name__"):
        name = dtype.__name__
    else:
        text = str(dtype)
        if text.startswith("<class '") and text.endswith("'>"):
            text = text.split("'")[1]
        if text.startswith("dtype(") and text.endswith(")"):
            inner = text[6:-1].strip("'\" ")
            text = inner or text
        name = text
    name = name.split(".")[-1]
    return name.strip(" '\"<>").lower()


def convert_dtype(
    dtype: Any | str | None,
    target_xp: Any,
    *,
    source_xp: Any | None = None,
) -> Any | None:
    """Convert a dtype between array API namespaces.

    Parameters
    ----------
    dtype : Any | str | None
        The dtype to convert. Can be a dtype object, string, or None.
    target_xp : module
        The target array API namespace to convert the dtype into.
    source_xp : module, optional
        The source namespace of the dtype. Provided for API symmetry and future
        use; currently unused but accepted.

    Returns
    -------
    Any | None
        The dtype object compatible with ``target_xp`` (or None if ``dtype`` is None).
    """
    if dtype is None:
        return None
    if target_xp is None:
        raise ValueError("target_xp must be provided to convert dtype.")

    target_name = getattr(target_xp, "__name__", "")
    dtype_module = getattr(dtype, "__module__", "")
    if dtype_module.startswith(target_name):
        return dtype
    if is_torch_namespace(target_xp) and str(dtype).startswith("torch."):
        return dtype

    name = _dtype_to_name(dtype)
    if not name:
        raise ValueError(f"Could not infer dtype name from {dtype!r}")

    candidates = dict.fromkeys(
        [name, name.lower(), name.upper(), name.capitalize()]
    )
    last_error: Exception | None = None
    for candidate in candidates:
        try:
            return resolve_dtype(candidate, target_xp)
        except ValueError as exc:
            last_error = exc

    # Fallback to direct attribute lookup
    attr = getattr(target_xp, name, None) or getattr(
        target_xp, name.lower(), None
    )
    if attr is not None:
        return attr

    raise ValueError(
        f"Unable to convert dtype {dtype!r} to namespace {target_name}"
    ) from last_error


def copy_array(x, xp: Any = None) -> Array:
    """Copy an array based on the array API being used.

    This uses the most appropriate method to copy the array
    depending on the array API.

    Parameters
    ----------
    x : Array
        The array to copy.
    xp : Any
        The array API to use for the copy.

    Returns
    -------
    Array
        The copied array.
    """
    if xp is None:
        xp = array_namespace(x)
    # torch does not play nicely since it complains about copying tensors
    if is_torch_namespace(xp):
        if is_torch_array(x):
            return xp.clone(x)
        else:
            return xp.as_tensor(x)
    else:
        try:
            return xp.copy(x)
        except (AttributeError, TypeError):
            # Fallback for array APIs that do not have a copy method
            return xp.array(x, copy=True)


def effective_sample_size(log_w: Array) -> float:
    xp = array_namespace(log_w)
    return xp.exp(xp.asarray(logsumexp(log_w) * 2 - logsumexp(log_w * 2)))


@contextmanager
def disable_gradients(xp, inference: bool = True):
    """Disable gradients for a specific array API.

    Usage:

    ```python
    with disable_gradients(xp):
        # Do something
    ```

    Parameters
    ----------
    xp : module
        The array API module to use.
    inference : bool, optional
        When using PyTorch, set to True to enable inference mode.
    """
    if is_torch_namespace(xp):
        if inference:
            with xp.inference_mode():
                yield
        else:
            with xp.no_grad():
                yield
    else:
        yield


def encode_dtype(xp, dtype):
    """Encode a dtype for storage in an HDF5 file.

    Parameters
    ----------
    xp : module
        The array API module to use.
    dtype : dtype
        The dtype to encode.

    Returns
    -------
    str
        The encoded dtype.
    """
    if dtype is None:
        return None
    return {
        "__dtype__": True,
        "xp": xp.__name__,
        "dtype": _dtype_to_name(dtype),
    }


def decode_dtype(xp, encoded_dtype):
    """Decode a dtype from an HDF5 file.

    Parameters
    ----------
    xp : module
        The array API module to use.
    encoded_dtype : dict
        The encoded dtype.

    Returns
    -------
    dtype
        The decoded dtype.
    """
    if isinstance(encoded_dtype, dict) and encoded_dtype.get("__dtype__"):
        if encoded_dtype["xp"] != xp.__name__:
            raise ValueError(
                f"Encoded dtype xp {encoded_dtype['xp']} does not match "
                f"current xp {xp.__name__}"
            )
        if is_torch_namespace(xp):
            return getattr(xp, encoded_dtype["dtype"].split(".")[-1])
        else:
            return xp.dtype(encoded_dtype["dtype"].split(".")[-1])
    else:
        return encoded_dtype


def encode_for_hdf5(value: Any) -> Any:
    """Encode a value for storage in an HDF5 file.

    Special cases:
    - None is replaced with "__none__"
    - Empty dictionaries are replaced with "__empty_dict__"
    """
    if is_jax_array(value) or is_torch_array(value):
        return to_numpy(value)
    if isinstance(value, CallHistory):
        return value.to_dict(list_to_dict=True)
    if isinstance(value, np.ndarray):
        return value
    if isinstance(value, (int, float, str)):
        return value
    if isinstance(value, (list, tuple)):
        if all(isinstance(v, str) for v in value):
            dt = h5py.string_dtype(encoding="utf-8")
            return np.array(value, dtype=dt)
        return [encode_for_hdf5(v) for v in value]
    if isinstance(value, set):
        return {encode_for_hdf5(v) for v in value}
    if isinstance(value, dict):
        if not value:
            return "__empty_dict__"
        else:
            return {k: encode_for_hdf5(v) for k, v in value.items()}
    if value is None:
        return "__none__"

    return value


def decode_from_hdf5(value: Any) -> Any:
    """Decode a value loaded from an HDF5 file, reversing encode_for_hdf5."""
    if isinstance(value, bytes):  # HDF5 may store strings as bytes
        value = value.decode("utf-8")

    if isinstance(value, str):
        if value == "__none__":
            return None
        if value == "__empty_dict__":
            return {}

    if isinstance(value, np.ndarray):
        # Try to collapse 0-D arrays into scalars
        if value.shape == ():
            return value.item()
        if value.dtype.kind in {"S", "O"}:
            try:
                return value.astype(str).tolist()
            except Exception:
                # fallback: leave as ndarray
                return value
        return value

    if isinstance(value, list):
        return [decode_from_hdf5(v) for v in value]
    if isinstance(value, tuple):
        return tuple(decode_from_hdf5(v) for v in value)
    if isinstance(value, set):
        return {decode_from_hdf5(v) for v in value}
    if isinstance(value, dict):
        return {
            k.decode("utf-8"): decode_from_hdf5(v) for k, v in value.items()
        }

    # Fallback for ints, floats, strs, etc.
    return value


def recursively_save_to_h5_file(h5_file, path, dictionary):
    """Save a dictionary to an HDF5 file with flattened keys under a given group path."""
    # Ensure the group exists (or open it if already present)
    group = h5_file.require_group(path)

    def _save_flattened(g, prefix, d):
        for key, value in d.items():
            full_key = f"{prefix}.{key}" if prefix else key
            if isinstance(value, dict):
                _save_flattened(g, full_key, value)
            else:
                try:
                    g.create_dataset(full_key, data=encode_for_hdf5(value))
                except TypeError as error:
                    try:
                        # Try saving as a string
                        dt = h5py.string_dtype(encoding="utf-8")
                        g.create_dataset(
                            full_key, data=np.array(str(value), dtype=dt)
                        )
                    except Exception:
                        raise RuntimeError(
                            f"Cannot save key {full_key} with value {value} to HDF5 file."
                        ) from error

    _save_flattened(group, "", dictionary)


def load_from_h5_file(h5_file, path):
    """Load a flattened dictionary from an HDF5 group and rebuild nesting."""
    group = h5_file[path]
    result = {}

    for key, dataset in group.items():
        parts = key.split(".")
        d = result
        for part in parts[:-1]:
            d = d.setdefault(part, {})
        d[parts[-1]] = decode_from_hdf5(dataset[()])

    return result


def get_package_version(package_name: str) -> str:
    """Get the version of a package.

    Parameters
    ----------
    package_name : str
        The name of the package.

    Returns
    -------
    str
        The version of the package.
    """
    try:
        module = __import__(package_name)
        return module.__version__
    except ImportError:
        return "not installed"


class AspireFile(h5py.File):
    """A subclass of h5py.File that adds metadata to the file."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._set_aspire_metadata()

    def _set_aspire_metadata(self):
        from . import __version__ as aspire_version

        if self.mode in {"w", "w-", "a", "r+"}:
            self.attrs["aspire_version"] = aspire_version
        else:
            aspire_version = self.attrs.get("aspire_version", "unknown")
            if aspire_version != "unknown":
                logger.warning(
                    f"Opened Aspire file created with version {aspire_version}. "
                    f"Current version is {aspire_version}."
                )


def update_at_indices(x: Array, slc: Array, y: Array) -> Array:
    """Update an array at specific indices."

    This is a workaround for the fact that array API does not support
    advanced indexing with all backends.

    Examples
    --------
    >>> x = xp.array([[1, 2], [3, 4], [5, 6]])
    >>> update_at_indices(x, (slice(None), 0), xp.array([10, 20, 30]))
    [[10  2]
     [20  4]
     [30  6]]

    Parameters
    ----------
    x : Array
        The array to update.
    slc : Array
        The indices to update.
    y : Array
        The values to set at the indices.

    Returns
    -------
    Array
        The updated array.
    """
    try:
        x[slc] = y
        return x
    except TypeError:
        return x.at[slc].set(y)


@dataclass
class CallHistory:
    """Class to store the history of calls to a function.

    Attributes
    ----------
    args : list[tuple]
        The positional arguments of each call.
    kwargs : list[dict]
        The keyword arguments of each call.
    """

    args: list[tuple]
    kwargs: list[dict]

    def to_dict(self, list_to_dict: bool = False) -> dict[str, Any]:
        """Convert the call history to a dictionary.

        Parameters
        ----------
        list_to_dict : bool
            If True, convert the lists of args and kwargs to dictionaries
            with string keys. If False, keep them as lists. This is useful
            when encoding the history for HDF5.
        """
        if list_to_dict:
            return {
                "args": {str(i): v for i, v in enumerate(self.args)},
                "kwargs": {str(i): v for i, v in enumerate(self.kwargs)},
            }
        else:
            return {
                "args": [list(arg) for arg in self.args],
                "kwargs": [dict(kwarg) for kwarg in self.kwargs],
            }


def track_calls(wrapped=None):
    """Decorator to track calls to a function.

    The decorator adds a :code:`calls` attribute to the wrapped function,
    which is a :py:class:`CallHistory` object that stores the arguments and
    keyword arguments of each call.
    """

    @wrapt.decorator
    def wrapper(wrapped_func, instance, args, kwargs):
        # If instance is provided, we're dealing with a method.
        if instance:
            # Attach `calls` attribute to the method's `__func__`, which is the original function
            if not hasattr(wrapped_func.__func__, "calls"):
                wrapped_func.__func__.calls = CallHistory([], [])
            wrapped_func.__func__.calls.args.append(args)
            wrapped_func.__func__.calls.kwargs.append(kwargs)
        else:
            # For standalone functions, attach `calls` directly to the function
            if not hasattr(wrapped_func, "calls"):
                wrapped_func.calls = CallHistory([], [])
            wrapped_func.calls.args.append(args)
            wrapped_func.calls.kwargs.append(kwargs)

        # Call the original wrapped function
        return wrapped_func(*args, **kwargs)

    return wrapper(wrapped) if wrapped else wrapper
